import numpy as np
import pytest

from numpy.testing import assert_almost_equal

from pykingenie.utils.fitting_surface import (
    guess_initial_signal,
    fit_one_site_dissociation
)

def test_guess_initial_signal_exception():

    assoc_time_lst = [[10]]
    assoc_signal_lst = [[10]]

    s0s = guess_initial_signal(assoc_time_lst,assoc_signal_lst)

    assert s0s[0] == 10, "The initial signal should be the same as the signal at the first time point"

def test_fit_one_site_dissociation_fixed_s0():

    t = np.linspace(0, 10, 50)
    s0 = 1.0
    k_off = 0.1

    y = s0 * np.exp(-k_off*t)

    initial_parameters = [0.5]

    low_bounds = [0.005]
    high_bounds = [10]

    fit_params, _, _ = fit_one_site_dissociation(
        [y], [t], initial_parameters, low_bounds, high_bounds,
        fit_s0=False)

    assert_almost_equal(fit_params[0],k_off,decimal=2)


import numpy as np
import pytest
import os

from pykingenie.kingenie_solution import KinGenieCsvSolution

kingenie = KinGenieCsvSolution('test_kingenie_csv')

test_file = "./test_files/kingenie_solution.csv"
test_file_2 = "./test_files/two_columns_solution.csv"
test_file_3 = "./test_files/three_columns_solution.csv"

test_file_error = "./test_files/one_column_solution.csv"

def test_instance_creation():
    """
    Test the creation of a KinGenieCsv instance.
    """
    assert isinstance(kingenie, KinGenieCsvSolution), "The instance should be of type KinGenieCsv."
    assert kingenie.name == "test_kingenie_csv"

def test_read_csv():

    kingenie.read_csv(test_file)

    # check kingenie.xs has some data
    assert len(kingenie.xs) > 0, "The xs list should not be empty after reading the CSV file."

def test_read_two_columns_csv():

    kingenie.read_csv(test_file_2)

    assert len(kingenie.xs) > 0, "The xs list should not be empty after reading the CSV file."

def test_read_three_columns_csv():

    kingenie.read_csv(test_file_3)

    assert len(kingenie.xs) > 0, "The xs list should not be empty after reading the CSV file."

def test_read_csv_error():

    with pytest.raises(ValueError):
        kingenie.read_csv(test_file_error)


from .surface_exp      import SurfaceBasedExperiment
import os
import json
import zipfile

import numpy as np
import pandas as pd

from .utils.processing import find_loading_column

class GatorExperiment(SurfaceBasedExperiment):
    """
    Class to handle Gator Bio-Layer Interferometry (BLI) experiments.
    
    This class extends the SurfaceBasedExperiment to provide functionality
    specific to Gator BLI instruments, including parsing experiment files
    and loading sensor data.
    
    Parameters
    ----------
    name : str, optional
        Name of the experiment, default is 'Gator_experiment'.
    """

    def __init__(self, name = 'Gator_experiment'):

        super().__init__(name, 'Gator_experiment')

    def read_all_gator_data(self, files, names=None):
        """
        Read all Gator data files for the experiment.
        
        This method reads all necessary files for a Gator experiment, including
        the experiment settings, step information, and sensor data.
        
        Parameters
        ----------
        files : str or list
            Path to a zip file containing Gator data files or a list of file paths.
            If a zip file is provided, it will be extracted to the same directory.
        names : list, optional
            List of file names corresponding to the file paths in `files`.
            If None, the file names will be extracted from the file paths.
            
        Returns
        -------
        None
            Updates instance attributes with the loaded experiment data.
        """

        # Check if files is a string
        if isinstance(files, str):
            # Check if files is a zipfile and extract it
            if files.endswith('.zip'):
                with zipfile.ZipFile(files, 'r') as zip_ref:
                    zip_ref.extractall(os.path.dirname(files))
                files = [os.path.join(os.path.dirname(files), file) for file in zip_ref.namelist()]

        self.traces_loaded = False

        if names is None:
            names = [os.path.basename(file) for file in files]

        # Find the file ExperimentStep.ini and read it
        for file, name in zip(files, names):
            if 'ExperimentStep.ini' in name:
                self.read_experiment_ini(file)

        # Find the file Setting.ini and read it
        for file, name in zip(files, names):
            if 'Setting.ini' in name:
                self.read_settings_ini(file)

        # Find the sensor csvs and read them
        self.read_sensor_data(files, names)

        return None

    def read_experiment_ini(self, file, nr_max_columns=12):
        """
        Read the experiment initialization file.
        
        Parses the ExperimentStep.ini file that contains information about the experimental steps,
        their durations, and types (baseline, association, dissociation, etc.).
        
        Parameters
        ----------
        file : str
            Path to the ExperimentStep.ini file.
        nr_max_columns : int, optional
            Maximum number of columns in the plate, default is 12.
            
        Returns
        -------
        None
            Updates the df_steps attribute with a DataFrame containing step information.
            
        Notes
        -----
        Example format for ExperimentStep.ini:
        
            [Experiment]
            Num=4
            [Experiment1]
            Step1=0,0,60,0
            Step2=1,60,104,1
            Step3=5,104,164,0
            Step4=6,164,464,2
            Step5=7,464,764,3
            Num=5
            ProbeIndex=12
            [ExperimentLog]
            Step1="{"StepType":0,"AssayNumber":1,"PlateIndex":1,"PlateColumnType":12,"RowLocation":0,"Location":0,"Speed":400,"Time":600,"SumTime":0,"ProbeIndex":0,"PickerUsed":1,"AllChannelUsed":[255],"ChannelUsed":0,"bAssayFirstStep":false,"kst":0,"Status":0}"
            Step2="{"StepType":3,"AssayNumber":1,"PlateIndex":1,"PlateColumnType":12,"RowLocation":0,"Location":1,"Speed":1000,"Time":5,"SumTime":0,"ProbeIndex":0,"PickerUsed":1,"AllChannelUsed":[255],"ChannelUsed":0,"bAssayFirstStep":false,"kst":0,"Status":0}"
            Step3="{"StepType":4,"AssayNumber":1,"PlateIndex":1,"PlateColumnType":12,"RowLocation":0,"Location":2,"Speed":1000,"Time":5,"SumTime":0,"ProbeIndex":0,"PickerUsed":1,"AllChannelUsed":[255],"ChannelUsed":0,"bAssayFirstStep":false,"kst":0,"Status":0}"
            Step4="{"StepType":3,"AssayNumber":1,"PlateIndex":1,"PlateColumnType":12,"RowLocation":0,"Location":1,"Speed":1000,"Time":5,"SumTime":0,"ProbeIndex":0,"PickerUsed":1,"AllChannelUsed":[255],"ChannelUsed":0,"bAssayFirstStep":false,"kst":0,"Status":0}"
            Step5="{"StepType":4,"AssayNumber":1,"PlateIndex":1,"PlateColumnType":12,"RowLocation":0,"Location":2,"Speed":1000,"Time":5,"SumTime":0,"ProbeIndex":0,"PickerUsed":1,"AllChannelUsed":[255],"ChannelUsed":0,"bAssayFirstStep":false,"kst":0,"Status":0}"
            Step6="{"StepType":3,"AssayNumber":1,"PlateIndex":1,"PlateColumnType":12,"RowLocation":0,"Location":1,"Speed":1000,"Time":5,"SumTime":0,"ProbeIndex":0,"PickerUsed":1,"AllChannelUsed":[255],"ChannelUsed":0,"bAssayFirstStep":false,"kst":0,"Status":0}"
            Step7="{"StepType":4,"AssayNumber":1,"PlateIndex":1,"PlateColumnType":12,"RowLocation":0,"Location":2,"Speed":1000,"Time":5,"SumTime":0,"ProbeIndex":0,"PickerUsed":1,"AllChannelUsed":[255],"ChannelUsed":0,"bAssayFirstStep":false,"kst":0,"Status":0}"
            Step8="{"StepType":2,"AssayNumber":1,"PlateIndex":0,"PlateColumnType":12,"RowLocation":0,"Location":12,"Speed":1000,"Time":60,"SumTime":0,"ProbeIndex":0,"PickerUsed":1,"AllChannelUsed":[255],"ChannelUsed":0,"bAssayFirstStep":true,"kst":0,"Status":0}"
            Step9="{"StepType":2,"AssayNumber":1,"PlateIndex":0,"PlateColumnType":12,"RowLocation":0,"Location":13,"Speed":400,"Time":44,"SumTime":60,"ProbeIndex":0,"PickerUsed":1,"AllChannelUsed":[255],"ChannelUsed":0,"bAssayFirstStep":false,"kst":1,"Status":1}"

        """

        with open(file, 'r') as f:
            lines = f.read().splitlines()

            # Find the line [ExperimentLog]
            for i, line in enumerate(lines):
                if line.startswith('[ExperimentLog]'):
                    start_index = i + 1
                    break

            # Count the number of items of the next line, when stripping by commas
            nr_items = len(lines[start_index].split(','))

            step_strings = []
            # Find all lines with the same amount of items and store them
            for line in lines[start_index:]:
                if len(line.split(',')) == nr_items:
                    step_strings.append(line)

        # List to hold dictionaries
        step_dicts = []

        # Iterate over the list of strings
        for step_string in step_strings:
            # Extract the JSON part of the string
            json_part = step_string.split('=', 1)[1].strip('"')
            # Convert the JSON string into a dictionary
            step_dict = json.loads(json_part)
            # Append the dictionary to the list
            step_dicts.append(step_dict)

        # Convert the list of dictionaries into a DataFrame
        df = pd.DataFrame(step_dicts)

        # Remove extra columns
        df = df[['kst', 'Location', 'Time', 'SumTime', 'AssayNumber', 'Speed']]
        # Add the column with the stepnumber to the beginning
        df.insert(0, 'StepNumber', np.arange(len(df)))

        # Change column name 'Location' to 'Column'
        df.rename(columns={'Location': 'Column'}, inplace=True)

        # Subtract 11 from it - beacuse the plate with the sample usually starts at index 12 to 24
        df['Column'] = df['Column'] - nr_max_columns + 1

        # If kst equals 2, then the step is a association
        # If kst equals 3, then the step is a dissociation

        # Guess if the step was 'ASSOC' or 'DISASSOC',
        # depending on the value of kst
        df['kst'] = df['kst'].replace({2: 'ASSOC', 3: 'DISASSOC', 1: 'LOADING', 0: 'OTHER'})

        # Change column name 'kst' to 'Type'
        df.rename(columns={'kst': 'Type'}, inplace=True)

        # Update the 'Type' column to 'BASELINE' for rows preceding 'ASSOC'
        df.loc[df['Type'].shift(-1) == 'ASSOC', 'Type'] = 'BASELINE'

        self.df_steps = df

        return None

    def read_settings_ini(self, file, nr_max_sensors=8, nr_max_columns=12):
        """
        Read the settings initialization file.
        
        Parses the Setting.ini file that contains information about the experiment settings,
        including sensor configurations, sample concentrations, and plate layout.
        
        Parameters
        ----------
        file : str
            Path to the Setting.ini file.
        nr_max_sensors : int, optional
            Maximum number of sensors per column, default is 8.
        nr_max_columns : int, optional
            Maximum number of columns in the plate, default is 12.
            
        Returns
        -------
        None
            Updates several instance attributes:
            - sensor_names: List of sensor names
            - ligand_conc_df: DataFrame with ligand concentration information
            
        Notes
        -----
        Example format for Setting.ini:
        
            [BasicInformation]
            PreExperiment="{"AssayDescription":"","AssayUser":"Marianna","CreationTime":"03-31-2025 14:01:50",...}"
            ExperimentType=Kinetics
            SaveTime=03/31/2025 15:55:27
            Frequency=10Hz
            DataVersion=2.15.5.1221
            [Regeneration]
            RegenerationEnable=1
            Repeat=3
            RTime=5
            RSpeed=1000
            NTime=5
            NSpeed=1000
            [SampleKinetics]
            Well1=4,-1,-1,-1,,
            Well2=4,-1,-1,-1,,
            ...
        """

        with open(file, 'r') as f:
            lines = f.read().splitlines()

            # Find the molar concentration units
            # below the line MolarConcentrationUnit=NM
            for line in lines:
                if line.startswith('MolarConcentrationUnit='):
                    units = line.split('=')[1].strip().lower()
                    factor = {'nm': 1e-3, 'mm': 1e3, 'pm': 1e-6, 'm': 1e6}.get(units, 1)
                    break

            # Find the line [SampleKinetics]
            for i, line in enumerate(lines):
                if line.startswith('[SampleKinetics]'):
                    start_index = i + 1
                    break

            # Count the number of items of the next line, when stripping by commas
            nr_items = len(lines[start_index].split(','))

            sample_strings = []
            # Find all lines with the same amount of items and store them
            for line in lines[start_index:]:
                if len(line.split(',')) == nr_items:
                    sample_strings.append(line)

            # Leave lines starting with 'Well'
            sample_strings = [line for line in sample_strings if line.startswith('Well')]

            # Now we need to convert wells that go from 1 to 96 into a combination of letters and numbers
            # They are ordered by rows, each column has nr_max_sensors rows
            concs = []

            locations = np.concatenate([np.repeat(x, nr_max_sensors)     for x in range(1, nr_max_columns+1)], axis=0)
            sensors   = np.concatenate([(65 + np.arange(nr_max_sensors)) for _ in range(1, nr_max_columns+1)], axis=0)
            # Apply chr function to sensors
            sensors = [chr(x) for x in sensors]

            labels = []

            for string in sample_strings:
                # We need to extract the concentration : Well49=1,17,170,100,,EGFR
                # We need to split by commas and take the fourth element
                conc = string.split('=')[1].split(',')[3]
                conc = float(conc) if conc != '' else 0

                concs.append(conc)

                # Extract the label, last column
                label = string.split(',')[-1]
                labels.append(label)

            # Create the ligand concentration dataframe
            df_sensor_prev = pd.DataFrame({'Sensor': sensors,
                                           'Location': locations,
                                           'Concentration_micromolar': concs,
                                           'SampleID': labels})

            loading_labels = []
            loading_columns = []
            sample_labels = []
            location = []
            sensor = []
            concs = []

            loading_columns_extended = []

            # Iterate over the associations and find loading column

            for row_index in range(len(self.df_steps)):

                step_type = self.df_steps.iat[row_index, self.df_steps.columns.get_loc('Type')]
                if step_type == 'ASSOC':

                    loading_column = find_loading_column(self.df_steps, row_index)

                    df_temp = df_sensor_prev[df_sensor_prev['Location'].isin([loading_column])]
                    loading_labels += df_temp['SampleID'].to_list()

                    loading_columns.append(loading_column)

                    association_column = self.df_steps.iat[row_index, self.df_steps.columns.get_loc('Column')]

                    location += [association_column for _ in range(nr_max_sensors)]
                    # The expression chr(65 + i) returns the uppercase ASCII letter corresponding to the index i,
                    # where i = 0 gives 'A', i = 1 gives 'B', and so on. For example, chr(65 + 2) returns 'C'.
                    sensor   += [chr(65 + i) for i in range(nr_max_sensors)]

                    loading_columns_extended += [loading_column for _ in range(nr_max_sensors)]

                    df_temp = df_sensor_prev[df_sensor_prev['Location'].isin([association_column])]
                    sample_labels += df_temp['SampleID'].to_list()

                    concs += df_temp['Concentration_micromolar'].to_list()
                else:
                    loading_columns.append('NA')

            # Include the loading column in df_steps
            self.df_steps['Loading_location'] = loading_columns

            sample_id = [x + '-' + y for x, y in zip(sample_labels, loading_labels)]

            replicates = [1 for _ in range(len(sample_id))]

            unq_sample_id = list(set(sample_id))

            for unq in unq_sample_id:

                # Find the index with that sample id
                idx = [i for i, x in enumerate(sample_id) if x == unq]

                replicates[idx[0]] = 1

                idx_cnt = 1
                for i in idx[1:]:
                    if sample_id[i] == sample_id[i - 1]:
                        idx_cnt += 1
                    replicates[i] = (idx_cnt - 1) // nr_max_sensors + 1

            # Create the ligand concentration dataframe
            df_sensor = pd.DataFrame({'Sensor': sensor,
                                      'Analyte_location': location,
                                      'Concentration_micromolar': concs,
                                      'Loading_location': loading_columns_extended,
                                      'SampleID': sample_id,
                                      'Replicate': replicates,
                                      'Experiment': self.name})

            # Multiply concentrations by the factor
            df_sensor['Concentration_micromolar'] = df_sensor['Concentration_micromolar'] * factor

            sensor_names = df_sensor['Sensor'].unique().tolist()
            self.sensor_names = sensor_names

            self.create_unique_sensor_names()

            self.ligand_conc_df = df_sensor

        return None

    def read_sensor_data(self, files, names=None):
        """
        Read the sensor data from CSV files.
        
        Parses the Gator sensor data CSV files containing time series data for each sensor channel.
        The method organizes the data according to experimental steps and sensors.
        
        Parameters
        ----------
        files : list
            List of file paths that may include sensor data CSV files.
        names : list, optional
            List of file names corresponding to the file paths in `files`.
            If None, the file names will be extracted from the file paths.
            
        Returns
        -------
        None
            Updates several instance attributes:
            - xs: List of time arrays for each sensor
            - ys: List of signal arrays for each sensor
            - df_steps: DataFrame with filtered step information containing only steps with data
            - no_steps: Number of steps with data
            - traces_loaded: Boolean indicating if traces were successfully loaded
            
        Notes
        -----
        The sensor data files are expected to be in CSV format with the following characteristics:
        - File names contain 'Channel' and have a '.csv' extension
        - The first column contains signal values
        - The second column contains time values
        """

        fns   = [fn for fn, name in zip(files, names) if '.csv' in name and 'Channel' in name]
        names = [name for name in names if '.csv' in name and 'Channel' in name]

        self.fns = fns

        # Initialize dictionaries with data
        xs = [[] for _ in range(len(self.sensor_names))]
        ys = [[] for _ in range(len(self.sensor_names))]

        # Find steps with data, where the SumTime is zero,
        # or if not zero, it is the same as the previous step

        steps_with_data = []

        for i in range(1, len(self.df_steps)):

            sum_time_i = self.df_steps.iat[i, self.df_steps.columns.get_loc('SumTime')]
            sum_time_prev = self.df_steps.iat[i - 1, self.df_steps.columns.get_loc('SumTime')]

            if sum_time_i == 0 or sum_time_i == sum_time_prev:
                continue
            else:
                steps_with_data.append(i - 1)

        # Create a copy of self.df_steps that will contain all the data and remove steps with no data
        df_steps_all = self.df_steps.copy()

        # Select the steps with data only
        self.df_steps = self.df_steps.iloc[steps_with_data]

        # Remove the stepnumber column
        self.df_steps = self.df_steps.drop(columns=['StepNumber'])

        # Reset the stepnumber column
        self.df_steps['StepNumber'] = np.arange(len(self.df_steps))

        # Reset the index of step_info_with_data
        self.df_steps = self.df_steps.reset_index(drop=True)

        # Rename the column 'Column' to 'Column_location' in df_steps
        self.df_steps.rename(columns={'Column': 'Column_location'}, inplace=True)

        # Move the column StepNumber to the first position
        column_to_move = self.df_steps.pop('StepNumber')
        self.df_steps.insert(0, 'StepNumber', column_to_move)

        # Move the column Loading_location to the fifth position
        column_to_move = self.df_steps.pop('Loading_location')
        self.df_steps.insert(4, 'Loading_location', column_to_move)

        # Convert the columns Loading_location and Column_location to string, for compatibility with get_step_xy
        self.df_steps['Loading_location'] = self.df_steps['Loading_location'].astype(str)
        self.df_steps['Column_location']  = self.df_steps['Column_location'].astype(str)

        self.no_steps = len(self.df_steps)

        # xs will have one element per sensor
        # each element will have as many subelements as steps

        # Extract the assay number and channel number from the file names
        # and sort the files by assay number and channel number
        assay_number = [int(name.split('_')[1].split('Channel')[0]) for name in names]
        channel_number = [int(name.split('Channel')[1].split('.')[0]) for name in names]

        # Find the indices that would sort the arrays by assay number and channel number
        sorted_indices = np.lexsort((channel_number, assay_number))
        # Sort the file names and assay numbers
        fns   = [fns[i] for i in sorted_indices]
        names = [names[i] for i in sorted_indices]

        for fn, name in zip(fns, names):

            try:

                assay_number_i   = int(name.split('_')[1].split('Channel')[0])
                channel_number_i = int(name.split('Channel')[1].split('.')[0])

                df = pd.read_csv(fn, skiprows=1, header=None)

                x = df.iloc[:, 1].to_numpy(dtype=float)
                y = df.iloc[:, 0].to_numpy(dtype=float)

                # Subset df_steps_all according to the assay number
                temp_df = df_steps_all[df_steps_all['AssayNumber'] == assay_number_i]

                # Filter which step numbers are in step_with_data
                temp_df = temp_df[temp_df['StepNumber'].isin(steps_with_data)]

                # Cumulative time for this assay
                temp_df['SumTime2'] = temp_df['Time'].cumsum()

                # Find starting sumtime for the assay
                start_sumtime = temp_df.iloc[0, temp_df.columns.get_loc('SumTime')]

                # For each step, include the data into xs and ys
                for i in range(len(temp_df)):

                    if i == 0:
                        start_time = 0
                    else:
                        start_time = temp_df.iloc[i - 1, temp_df.columns.get_loc('SumTime2')]

                    step_time = temp_df.iloc[i, temp_df.columns.get_loc('Time')]

                    end_time = start_time + step_time

                    sel_idx = np.logical_and(x >= start_time, x < end_time)

                    y_sel = y[sel_idx]
                    x_sel = x[sel_idx]

                    # Add the overall sumtime to x
                    x_sel += start_sumtime

                    # Append the second column to xs
                    xs[channel_number_i - 1].append(x_sel)

                    # Append the first column to ys
                    ys[channel_number_i - 1].append(y_sel)

                self.traces_loaded = True
                self.xs = xs
                self.ys = ys

            except:

                pass

        return None



import pandas as pd
import numpy  as np

from .solution_exp import SolutionBasedExp

class KinGenieCsvSolution(SolutionBasedExp):
    """
    A class used to represent a KinGenie simulation (solution-based).
    
    This class handles KinGenie solution simulation data that can be exported from the Simulation panel.

    Parameters
    ----------
    name : str
        Name of the experiment.

    Attributes
    ----------
    name : str
        Name of the experiment.
    xs : list of numpy.ndarray
        List of x values (time), one array per trace.
    ys : list of numpy.ndarray
        List of y values (signal), one array per trace.
    no_traces : int
        Number of traces.
    traces_names : list of str
        List of trace names, one per trace.
    traces_names_unique : list of str
        List of unique trace names.
    conc_df : pandas.DataFrame
        DataFrame with the concentration information for each trace.
    traces_loaded : bool
        Whether traces have been loaded.
    """

    def __init__(self, name):

        super().__init__(name, 'kingenie_csv_solution')
        
    def read_csv(self, file):
        """
        Read the KinGenie CSV file containing solution-based simulation data.
        
        Parameters
        ----------
        file : str
            Path to the CSV file.
            
        Returns
        -------
        None
            The method populates class attributes with data from the CSV file.
            
        Notes
        -----
        The CSV file should have the following columns:
        - Time: time points of the simulation
        - Signal: signal values at each time point
        - Protein_concentration_micromolar: protein concentration
        - Ligand_concentration_micromolar: ligand concentration
        
        Example CSV format:
        
            Time    Signal  Protein_concentration_micromolar Ligand_concentration_micromolar
            0       0       5                                0.1
            0.5     1       5                                0.1
        
        After reading, this method creates the following attributes:
        - self.xs: list of time values for each trace
        - self.ys: list of signal values for each trace
        - self.no_traces: number of unique traces
        - self.traces_names: list of trace names
        - self.conc_df: DataFrame with concentration information
        """

        df = pd.read_csv(file)

        # Validate the input file
        # Check we have at least 2 columns, time versus signal

        if len(df.columns) < 2:
            raise ValueError("The CSV file must have at least 2 columns.")

        # If we only have two columns, assume that the third is the protein concentration
        if len(df.columns) == 2:
            df['Protein_concentration_micromolar'] = 0

        # If we have only 3 columns, assume that the third is the protein concentration, and the ligand concentration is 0
        if len(df.columns) == 3:
            df['Ligand_concentration_micromolar'] = 0

        # Assume the first 4 columns are Time, Signal, Protein_concentration_micromolar, Ligand_concentration_micromolar
        df = df.iloc[:, 0:4]

        df.columns = ['Time', 'Signal', 'Protein_concentration_micromolar', 'Ligand_concentration_micromolar']

        # Find the ligand concentrations
        ligand_conc = df['Ligand_concentration_micromolar']

        # Find the protein concentrations
        protein_conc = df['Protein_concentration_micromolar']

        # Combine protein and ligand concentration into one array
        combined_concs_array = [str(x) + ' and ' + str(y) for x, y in zip(protein_conc, ligand_conc)]

        # Add a new column to the dataframe
        df['Combined_concentration'] = combined_concs_array

        # Find the unique combined concentrations
        combined_concs_unq = np.unique(combined_concs_array)

        self.no_traces = len(combined_concs_unq)

        # Use fake names
        self.traces_names = ['trace ' + str(i + 1) for i in range(self.no_traces)]

        # Initiate self.xs and self.ys
        self.xs = []
        self.ys = []

        protein_conc_unqs = []
        ligand_conc_unqs  = []

        # Now populate, for each sensor self.xs and self.ys
        for i, cc in enumerate(combined_concs_unq):
            # Extract the rows with the same combined concentrations
            df_temp = df[df['Combined_concentration'] == cc]

            protein_conc_unqs.append(df_temp['Protein_concentration_micromolar'].to_numpy()[0])
            ligand_conc_unqs.append(df_temp['Ligand_concentration_micromolar'].to_numpy()[0])

            # Extract the time values of the association/dissociation phase
            time_int = df_temp['Time'].to_numpy()

            # Populate self.xs
            self.xs.append(time_int)

            # Extract the signal values
            signal = df_temp['Signal'].to_numpy()

            # Populate self.ys
            self.ys.append(signal)

        # Now generate a fake ligand concentration df
        df_traces = pd.DataFrame({
            'Trace': self.traces_names,
            'Protein_concentration_micromolar': protein_conc_unqs,
            'Ligand_concentration_micromolar': ligand_conc_unqs,
            'SampleID': 'simulation',
            'Experiment': self.name})

        self.conc_df = df_traces

        self.create_unique_traces_names()
        self.traces_loaded = True

        return None


import pytest
import numpy as np

from pykingenie.utils.signal_solution import (
    signal_ode_one_site_insolution,
    signal_ode_induced_fit_insolution,  # now the reduced version
    signal_ode_conformational_selection_insolution,
    get_initial_concentration_conformational_selection,
    get_kobs_induced_fit,
    get_kobs_conformational_selection
)

def test_solve_ode_one_site_insolution():

    t = np.linspace(0, 10, 10)
    koff = 0.1
    Kd = 0.5
    a_total = 1.0
    b_total = 1.0
    t0 = 0

    signal_a=0
    signal_b=0
    signal_complex=1

    result = signal_ode_one_site_insolution(t, koff, Kd, a_total, b_total, t0,signal_a, signal_b, signal_complex)

    expected = [0,0.17265383,0.2791352,0.34808851,0.39423549,0.42570054,0.44748987,0.46274577,0.47348593,0.48107122]

    np.testing.assert_array_almost_equal(result, expected, decimal=4, err_msg="The result of the ODE solution does not match the expected values.")

    # Check that the response increases with time
    assert np.all(np.diff(result) >= 0), "The response should be non-decreasing with increasing time."


def test_signal_ode_induced_fit_insolution():

    t = np.linspace(0, 10, 10)
    y = [0,0] # Initial concentrations of E and S
    k1 = 100
    k_minus1 = 100
    k2 = 10
    k_minus2 = 1

    E_tot = 0.5
    S_tot = 0.5

    result = signal_ode_induced_fit_insolution(
        t, y, k1, k_minus1, k2, k_minus2,
        E_tot, S_tot,
        t0=0, signal_E=0, signal_S=0, signal_ES_int=1, signal_ES=1)

    expected = np.array([0.    , 0.3234, 0.3274, 0.3275, 0.3275, 0.3275, 0.3275, 0.3275, 0.3275, 0.3275])

    np.testing.assert_array_almost_equal(result, expected, decimal=4, err_msg="The result of the ODE solution (induced fit) does not match the expected values.")

    # Check that the response increases with time
    assert np.all(np.diff(result) >= -1e-6), "The response should be non-decreasing with increasing time. Tolerance is set to -1e-6 to account for numerical precision issues."


def test_signal_ode_conformational_selection_insolution():

    k1 = 100
    k_minus1 = 1
    k2 = 10
    k_minus2 = 100

    E_tot = 1
    S_tot = 0.5

    E1, E2 = get_initial_concentration_conformational_selection(E_tot,k1,k_minus1)

    t = np.linspace(0, 0.05, 10)

    y = [E2,0]

    result = signal_ode_conformational_selection_insolution(
        t, y, k1, k_minus1, k2,k_minus2,
        E_tot, S_tot,
        t0=0,signal_E1=0,signal_E2=0,signal_S=0,signal_E2S=1)

    expected = np.array([0.    , 0.0203, 0.0311, 0.0368, 0.0398, 0.0414, 0.0423, 0.0428,0.043 , 0.0431])

    np.testing.assert_array_almost_equal(result, expected, decimal=4, err_msg="The result of the ODE solution (conformational selection) does not match the expected values.")

def test_get_kobs_induced_fit():

    tot_lig = 0.5
    tot_prot = 0.5
    kr = 1
    ke= 10
    kon = 100
    koff = 100
    dominant=True

    result = get_kobs_induced_fit(tot_lig, tot_prot, kr, ke, kon, koff, dominant)

    assert np.isclose(result, 10.38016, atol=1e-3), "The kobs value for induced fit does not match the expected value."

def test_get_kobs_conformational_selection():
    
    tot_lig = 0.5
    tot_prot = 0.5
    kr = 100
    ke= 10
    kon = 10
    koff = 10
    dominant= True

    result = get_kobs_conformational_selection(tot_lig, tot_prot, kr, ke, kon, koff, dominant)

    assert np.isclose(result, 10.4155, atol=1e-3), "The kobs value for conformational selection does not match the expected value."




import numpy as np
import pytest
import os

from pykingenie.octet import OctetExperiment

folder = "./test_files/"
frd_files = os.listdir(folder)

frd_files = [os.path.join(folder, file) for file in frd_files if file.endswith('.frd') and file.startswith('230309')]
fmf_file  = os.path.join(folder, '230309_ExpMethod.fmf')


def test_instance_creation():

    bli = OctetExperiment('test_octet')
    assert isinstance(bli, OctetExperiment), "The instance should be of type OctetExperiment."
    assert bli.name == "test_octet"

def test_import_str_file():

    bli = OctetExperiment('test_octet')
    bli.read_sensor_data(frd_files[0])

    assert len(bli.xs) > 0, "The xs array should not be empty after reading a single file."

def test_import_no_frd_files():

    bli = OctetExperiment('test_octet')
    bli.read_sensor_data(fmf_file)

    assert bli.xs is None, "The xs list should be still None when no .frd files are provided."

def test_different_metadata_value_err():

    err_file = './test_files/different_sensor_metadata.frd'
    bli = OctetExperiment('test_octet')

    with pytest.raises(ValueError):
        bli.read_sensor_data([err_file] + frd_files)

def test_import_file_with_no_loading_data():

    # Create an instance of OctetExperiment
    bli = OctetExperiment('test_octet')
    bli.read_sensor_data(frd_files)

    bli.df_steps.loc[bli.df_steps['Type'] == 'LOADING', 'Type'] = 'BASELINE'
    bli.generate_ligand_conc_df()

    assert len(bli.ligand_conc_df) > 0

def test_merge_consecutive_steps():

    # Create an instance of OctetExperiment
    bli = OctetExperiment('test_octet')

    bli.read_sensor_data(frd_files)

    x_ref      = bli.xs[0][2]
    x_to_merge = bli.xs[0][3]

    bli.merge_consecutive_steps(3,4)

    x_merged = bli.xs[0][2]

    assert all(np.concatenate((x_ref, x_to_merge)) == x_merged), "The consecutive steps should be merged correctly."

    # Test reading sensor data with a non-empty list (mocked)
    # This part would normally read files, but we can mock it for testing purposes
    # Here we assume that the method works correctly when provided with valid file paths

def test_merge_consecutive_steps_2():

    # Create an instance of OctetExperiment
    bli = OctetExperiment('test_octet')

    bli.read_sensor_data(frd_files)

    # modify t0 so t0 ref is higher than t0 to merge
    bli.xs[0][2] += 100

    x_ref      = bli.xs[0][2]
    x_to_merge = bli.xs[0][3]

    bli.merge_consecutive_steps(3,4)

    x_merged = bli.xs[0][2]

    assert all(np.concatenate((x_to_merge, x_ref)) == x_merged), "The consecutive steps should be merged correctly."

def test_merge_steps_by_name():

    # Create an instance of OctetExperiment
    bli = OctetExperiment('test_octet')

    bli.read_sensor_data(frd_files)

    # Merge consecutive steps by name
    bli.merge_consecutive_steps_by_name('Association')

    # Check if the xs list has been updated correctly
    assert len(bli.xs) > 0, "The xs list should not be empty after merging steps by name."

def test_read_sample_plate_info():

    # Create an instance of OctetExperiment
    bli = OctetExperiment('test_octet')

    bli.read_sensor_data(frd_files)

    # run read_sample_plate_info without a valid fmf file
    bli.read_sample_plate_info('test')

    assert not bli.sample_plate_loaded, "The sample plate should not be loaded when an invalid fmf file is provided."

    # Read the plate information
    bli.read_sample_plate_info(fmf_file)

    # Check if sample_row is not None
    assert bli.sample_row is not None, "The sample_row should not be None after reading plate info."

def test_subtract_experiment():
    
    # Create two instances of OctetExperiment
    bli1 = OctetExperiment('test_octet_1')
    bli2 = OctetExperiment('test_octet_2')

    bli1.read_sensor_data(frd_files[:4])
    bli2.read_sensor_data(frd_files[:4])

    bli1.subtract_experiment(bli2)

    assert len(bli1.ys) > 0, "The ys list should not be empty after subtraction."

    # The ys should be zero because both experiments are identical
    for step_data in bli1.ys:
        for sensor_data in step_data:
            assert all(sensor_data == 0), "The ys data should be zero after subtracting identical experiments."

def test_subtract_experiment_not_in_place():
    
    # Create two instances of OctetExperiment
    bli1 = OctetExperiment('test_octet_1')
    bli2 = OctetExperiment('test_octet_2')

    bli1.read_sensor_data(frd_files[:4])
    bli2.read_sensor_data(frd_files[:4])

    bli1.subtract_experiment(bli2, inplace=False)

    assert len(bli1.ys) > 0, "The ys list should not be empty after subtraction."

    # We should have 8 sensors now in bli1
    assert len(bli1.sensor_names) == 8, "The number of sensors should be doubled after subtraction in non-inplace mode."

    # The ys of the last four sensors should be zero because both experiments are identical
    for step_data in bli1.ys[-4:]:
        for sensor_data in step_data:
            assert all(sensor_data == 0), "The ys data should be zero after subtracting identical experiments."

    # The ys of the first four sensors should not be zero
    for step_data in bli1.ys[:4]:
        for sensor_data in step_data:
            assert not all(sensor_data == 0), "The ys data should not be zero for the original sensors."


def test_subtract_experiment_error_1():
    
    # Create two instances of OctetExperiment
    bli1 = OctetExperiment('test_octet_1')
    bli2 = OctetExperiment('test_octet_2')

    bli1.read_sensor_data(frd_files[:4])
    bli2.read_sensor_data(frd_files[:3])  # Different number of sensors to trigger error

    with pytest.raises(RuntimeError, match="Experiments have different number of sensors"):
        bli1.subtract_experiment(bli2)

def test_subtract_experiment_error_2():
    
    # Create two instances of OctetExperiment
    bli1 = OctetExperiment('test_octet_1')
    bli2 = OctetExperiment('test_octet_2')

    bli1.read_sensor_data(frd_files[:4])
    bli2.read_sensor_data(frd_files[:4])  

    bli2.xs[0][0] = bli2.xs[0][0] + 10  # Modify time data to trigger error

    with pytest.raises(RuntimeError, match="Experiments have different time data"):
        bli1.subtract_experiment(bli2)

def test_convert_to_numbers():

    # Create an instance of OctetExperiment
    bli = OctetExperiment('test_octet')

    bli.read_sensor_data(frd_files)

    # List of entries in step info
    entries = ['Concentration', 'MolarConcentration', 'MolecularWeight', 'Temperature', 'StartTime',
               'AssayTime', 'FlowRate', 'ActualTime', 'CycleTime']

    # Change type to trigger exception
    bli.step_info[0]['Concentration'] = ['hello', 'world']

    bli.convert_to_numbers()

    assert all(bli.step_info[0]['Concentration'] == np.array([-1,-1]))


import re
import base64
import xml.etree.ElementTree as elementTree

import numpy as np
import pandas as pd

from .utils.processing import etree_to_dict, combine_dicts
from .surface_exp      import SurfaceBasedExperiment

factor_conc_to_micro = {'nM':1e-3, 'µM':1, 'mM':1e3, 'M':1e6, 'mg/ml':1e3, 'µg/ml':1}

class OctetExperiment(SurfaceBasedExperiment):
    """
    OctetExperiment class for handling Octet BLI data.

    Parameters
    ----------
    name : str
        Name of the experiment.

    Attributes
    ----------

        name (str):                 name of the experiment
        fns (list):                 list of file names (length n, one per sensor)
        xs (list):                  list of x values (time, length n, one per sensor)
        ys (list):                  list of y values (length n, one per sensor)
        exp_info (list):            list of dictionaries with experimental information
        step_info (list):           list of dictionaries with step information
        no_steps (int):             number of steps
        no_sensors (int):           number of sensors
        sample_column (np.ndarray): array of sample column information (96 elements, one per well)
        sample_row (np.ndarray):    array of sample row information    (96 elements, one per well)
        sample_type (list):         list of sample types   (96 elements, one per well)
        sample_id (list):           list of sample ids     (96 elements, one per well)
        sensor_names_unique (list): list of unique sensor names (length n, one per sensor)
        sensor_names (list):        list of sensor names (length n, one per sensor)
        df_steps    (pd.DataFrame): dataframe with the steps information
        ligand_conc_df (pd.DataFrame): dataframe with the ligand concentration information

        ligand_conc_df.head(2):
        Sensor Analyte_location  Concentration_micromolar SampleID  Replicate Experiment
        A1        5                    0.1300       wt          1          t
        B1        5                    0.0692       wt          1          t

        traces_loaded (bool):       True if traces are loaded
        sample_plate_loaded (bool): True if sample plate information is loaded
        sample_conc (np.array):     array with the sample concentrations (96 elements, one per well)
        sample_conc_labeled (list): list with the sample concentrations labeled (96 elements, one per well)
        steps_performed (pd.DataFrame): dataframe with the steps performed

        steps_performed.head(2):
        #Step         Type              Column   Time
        1            Regeneration      1        25
        2  BaselineNeutralization      2        90
        3         BaselineLoading      3        50
    """

    def __init__(self, name = 'BLI_experiment'):
        """
        Initialize the OctetExperiment instance.
        
        Parameters
        ----------
        name : str, optional
            Name of the experiment. Default is 'BLI_experiment'.
        """

        super().__init__(name,'BLI_experiment')

    def read_sensor_data(self, files, names=None):
        """
        Read the sensor data from the .frd files.
        
        Parameters
        ----------
        files : str or list of str
            Path(s) to .frd file(s) to read.
        names : str or list of str, optional
            Name(s) to assign to the sensors. If None, file names are used.
            
        Returns
        -------
        None
            The method populates class attributes with data from the files.
            
        Notes
        -----
        This method creates the following attributes:
            
        - traces_loaded : bool
          Whether traces were successfully loaded.
        - xs : list
          List of x values (time) for each sensor.
        - ys : list
          List of y values (signal) for each sensor.
        - exp_info : list
          List of dictionaries with experimental information.
        - step_info : list
          List of dictionaries with step information.
        - no_steps : int
          Number of steps in the experiment.
        - no_sensors : int
          Number of sensors in the experiment.
        """

        if names is None:
            names = files

        if not isinstance(files, list):
            files = [files]
            names = [names]

        fns = [fn for fn,name in zip(files,names) if '.frd' in name]

        if len(fns) < 1:
            self.traces_loaded = False
            return None
        else:
            self.fns = fns

        # Initialize dictionaries with data
        xs, ys, all_expinfo, all_stepinfo, more_info = [], [], [], [], []
        for fn in fns:
            # Load file
            tree = elementTree.parse(fn)
            root = tree.getroot()

            # Extract experimental info
            all_expinfo.append(etree_to_dict(root.find('ExperimentInfo')))

            # Initialize lists for each file
            x_values, y_values, step_info = [], [], []
            more_dict = {'FlowRate': [], 'StepType': [], 'StepName':[], 'StepStatus':[], 'ActualTime':[], 'CycleTime':[]}
            for step in root.find('KineticsData'):
                for step_x in step.findall('AssayXData'):
                    # Convert string to binary
                    data_text = bytes(step_x.text, 'utf-8')
                    # Convert to base64
                    decoded = base64.decodebytes(data_text)
                    # And now convert to float32 array
                    data_values = np.array(np.frombuffer(decoded, dtype=np.float32))
                    x_values.append(data_values)
                for step_y in step.findall('AssayYData'):
                    # Convert string to binary
                    data_text = bytes(step_y.text, 'utf-8')
                    # Convert to base64
                    decoded = base64.decodebytes(data_text)
                    # And now convert to float32 array
                    data_values = np.array(np.frombuffer(decoded, dtype=np.float32))
                    y_values.append(data_values)
                for step_data in step.findall('CommonData'):
                    step_info.append(etree_to_dict(step_data))
                for tag in ['FlowRate', 'StepType', 'StepName', 'StepStatus', 'ActualTime', 'CycleTime']:
                    for step_data in step.findall(tag):
                        more_dict[tag].append(step_data.text)

            xs.append(x_values)
            ys.append(y_values)
            all_stepinfo.append(combine_dicts(step_info))
            more_info.append(more_dict)

        # Merge all_stepinfo and more_info
        for i in range(len(all_stepinfo)):
            all_stepinfo[i] = {**all_stepinfo[i], **more_info[i]}

        self.no_sensors = len(self.fns)

        if self.no_sensors > 1:

            # Perform a filter to verify that all steps of the different sensors have the same type and location
            reference_location = all_stepinfo[0]['SampleLocation']
            reference_type     = all_stepinfo[0]['StepType']

            for i in range(1,len(all_stepinfo)):

                condition1 = all_stepinfo[i]['SampleLocation'] == reference_location
                condition2 = all_stepinfo[i]['StepType']     == reference_type

                if not condition1 or not condition2:

                    raise ValueError("The steps of the different sensors do not have the same type and location. Please import in one experiment only sensors with the same metadata")

        # Fill instance
        self.xs   = xs
        self.ys   = ys
        self.exp_info  = all_expinfo
        self.step_info = all_stepinfo
        # Convert text to floats
        self.convert_to_numbers()

        self.generate_ligand_conc_df()

        return None

    def generate_ligand_conc_df(self):
        """
        Generate a DataFrame with the analyte concentrations based on step_info and exp_info.
        
        Parameters
        ----------
        None
            
        Returns
        -------
        None
            This method populates class attributes related to ligand concentration.
            
        Notes
        -----
        This method requires the following attributes to be already populated:
            
        - step_info : list
          List of dictionaries with step information.
        - exp_info : list
          List of dictionaries with experimental information.
        - fns : list
          List of file names.
            
        This method creates/updates the following attributes:
            
        - no_steps : int
          Number of steps in the experiment.
        - no_sensors : int
          Number of sensors.
        - sensor_names : list
          List of sensor names.
        - df_steps : pandas.DataFrame
          DataFrame with step information.
        - ligand_conc_df : pandas.DataFrame
          DataFrame with ligand concentration information.
        """

        self.no_steps = len(self.step_info[0]['ActualTime'])

        self.sensor_names = [self.exp_info[i]['SensorName'] for i in range(self.no_sensors)]

        steps_names = self.step_info[0]['StepName']
        steps_types = self.step_info[0]['StepType']
        steps_start = self.step_info[0]['StartTime'] / 1000 # To seconds
        steps_loc   = self.step_info[0]['SampleLocation']

        self.df_steps = pd.DataFrame({'#Step':np.arange(len(steps_names))+1,
                                      'Name':steps_names,
                                      'Type':steps_types,
                                      'Start':steps_start,
                                      'Column_location':steps_loc})

        # We need to include the loading location in self.df_steps
        loading_location = []
        for row in self.df_steps.iterrows():
            step_type = row[1]['Type']
            if step_type == 'ASSOC':
                # Find the previous loading step
                for i in range(row[0],0,-1):

                    if self.df_steps.iloc[i]['Type'] == 'LOADING':
                        loading_location.append(self.df_steps.iloc[i]['Column_location'])
                        break
                    # If no loading step is found, append NaN
                    if i == 1:
                        loading_location.append(np.nan)

            else:
                loading_location.append(np.nan)

        self.df_steps['Loading_location'] = loading_location

        sensor_locs_all = np.concatenate([self.step_info[i]['SampleLocation']       for i in range(self.no_sensors)])
        sensor_type_all = np.concatenate([self.step_info[i]['StepType']             for i in range(self.no_sensors)])

        sensor_molar_conc_all = np.concatenate([self.step_info[i]['MolarConcentration']  for i in range(self.no_sensors)])
        sensor_mass_conc_all  = np.concatenate([self.step_info[i]['Concentration']       for i in range(self.no_sensors)])

        sensor_conc_all = []

        for i in range(len(sensor_molar_conc_all)):

            if sensor_molar_conc_all[i] < 0:

                sensor_conc_all.append(sensor_mass_conc_all[i])

            else:

                sensor_conc_all.append(sensor_molar_conc_all[i])

        sensor_conc_all = np.array(sensor_conc_all)

        sample_id_all   = np.concatenate([self.step_info[i]['SampleID']   for i in range(self.no_sensors)])

        sensor_name_rep = np.concatenate([np.repeat(self.exp_info[i]['SensorName'],len(self.step_info[i]['Concentration'])) for i in range(self.no_sensors)])

        conc_units = np.concatenate([self.step_info[i]['MolarConcUnits'] for i in range(self.no_sensors)])

        df_all = pd.DataFrame({'Sensor':sensor_name_rep,
                               'Analyte_location':sensor_locs_all,
                               'Type':sensor_type_all,
                               'Concentration_micromolar':sensor_conc_all,
                               'ConcUnits':conc_units,
                               'SampleID':sample_id_all})

        # For each association step, find the corresponding loading step
        # and add the loading step to the dataframe as a column next to the association

        # Add empty column to the data frame
        loading_location  = []
        loading_sample_id = []

        for i in range(len(df_all)):

            row = df_all.iloc[i]
            # Find if row is association step
            if row['Type'] == 'ASSOC':

                # Find the previous loading step
                for j in range(i,0,-1):

                    if df_all.iloc[j]['Type'] == 'LOADING':

                        loading_location.append(df_all.iloc[j]['Analyte_location'])
                        loading_sample_id.append(df_all.iloc[j]['SampleID'])
                        break

        # Keep only association or dissociation steps
        df = df_all[df_all['Type'] == 'ASSOC'].copy()

        # If loading location is empty, fill with 0
        if len(loading_location) == 0:
            loading_location =  [0] * len(df)
            loading_sample_id = [0] * len(df)

        # ADD column loading_location
        df['Loading_location']  = loading_location

        # Replace None with empty string in loading_sample_id
        loading_sample_id = [x if x is not None else '' for x in loading_sample_id]

        # Include the loading_sample_id, if we have more than one unique value
        unq_loading_ids = np.unique(loading_sample_id)

        if len(unq_loading_ids) > 1:

            # Combine the sample id with the loading id
            df['SampleID'] = df['SampleID'] + ' - ' + loading_sample_id

        # Remove the Type column
        df = df.drop(columns=['Type'])

        # Sort by location and sensor name
        df = df.sort_values(by=['Loading_location','Analyte_location','Sensor'])

        # Add rep column
        sizes = df.groupby(['Loading_location','Analyte_location','Sensor']).size().reset_index(name="Repetitions")

        rep_number = []

        for i in range(len(sizes)):
            rep_number.extend(np.arange(sizes['Repetitions'].iloc[i])+1)

        # Group by sensor and location
        df['Replicate'] = rep_number

        # Sort the dataframe first by sensor, second by Loading location,
        # Third by replicate and finally by location

        df = df.sort_values(by=['Analyte_location','Replicate','Loading_location','Sensor'])

        df['Factor'] = df.apply(lambda x: factor_conc_to_micro[x['ConcUnits']], axis=1)

        df['Concentration_micromolar'] = df['Concentration_micromolar'] * df['Factor']

        # Remove the factor column and conc units
        df = df.drop(columns=['Factor','ConcUnits'])

        # Add the experiment name
        df['Experiment'] = self.name

        self.ligand_conc_df = df

        self.create_unique_sensor_names()

        self.traces_loaded = True

        return None

    def merge_consecutive_steps(self, idx_ref, idx_to_merge):
        """
        Combine two consecutive steps into one step.
        
        Parameters
        ----------
        idx_ref : int
            Index of the reference step. The type of step will be taken from this step.
        idx_to_merge : int
            Index of the step to merge with the reference step.
            
        Returns
        -------
        None
            The method modifies the xs, ys, and step information in place.
            
        Notes
        -----
        The two steps must be consecutive (their indices must differ by exactly 1).
        """

        assert np.abs(idx_ref - idx_to_merge) == 1, "The two steps must be consecutive"
        assert idx_ref != idx_to_merge, "The two steps must be different"

        idx_ref -= 1      # Adjust for 0-based indexing
        idx_to_merge -= 1 # Adjust for 0-based indexing

        for sensor in range(self.no_sensors):

            # Extract the reference step x and y values
            x_ref = self.xs[sensor][idx_ref]
            y_ref = self.ys[sensor][idx_ref]

            # Extract the step to merge x and y values
            x_merge = self.xs[sensor][idx_to_merge]
            y_merge = self.ys[sensor][idx_to_merge]

            # Find wich has the lowest start time
            t0_ref   = np.min(x_ref)
            t0_merge = np.min(x_merge)

            # Concatenate the x and y values
            if t0_ref < t0_merge:

                self.xs[sensor][idx_ref] = np.concatenate((x_ref, x_merge))
                self.ys[sensor][idx_ref] = np.concatenate((y_ref, y_merge))

            else:

                self.xs[sensor][idx_ref] = np.concatenate((x_merge, x_ref))
                self.ys[sensor][idx_ref] = np.concatenate((y_merge, y_ref))

            # Remove the merge step from the xs and ys lists
            self.xs[sensor].pop(idx_to_merge)
            self.ys[sensor].pop(idx_to_merge)

            # Remove the step from self.step_info
            # using the idx_to_merge index

            for sensor_id in range(self.no_sensors):

                step_info = self.step_info[sensor_id]

                for i,l in enumerate(step_info):

                    values = self.step_info[sensor_id][l]

                    if isinstance(values, list) and len(values) > idx_to_merge:
                        self.step_info[sensor_id][l].pop(idx_to_merge)
                    elif isinstance(values, np.ndarray) and len(values) > idx_to_merge:
                        self.step_info[sensor_id][l] = np.delete(values, idx_to_merge)

                #self.step_info[i].pop(key)

            # Create the analyte concentration dataframe again

            self.generate_ligand_conc_df()

            return None

    def merge_consecutive_steps_by_name(self, step_name, reference=True):
        """
        Merge steps with a specific name with their consecutive step.
        
        Parameters
        ----------
        step_name : str
            Name of the step to merge with the next/previous step.
        reference : bool, optional
            If True, the step with name step_name will be used as reference 
            to extract the analyte concentration, loading location, etc.
            Default is True.
            
        Returns
        -------
        None
            The method modifies the xs, ys, and step information in place.
            
        Notes
        -----
        This method finds all steps with the given name and merges them with 
        their adjacent step. The merged step inherits properties from the
        reference step.
        """
        # Find the indices of the steps of type step_type
        idxs = []
        for i in range(len(self.df_steps)):
            if self.df_steps.iloc[i]['Name'] == step_name:

                # check if idx+2 is valid
                if i+2 <= len(self.df_steps):
                    idxs.append(i)

        # Sort them in reverse order to avoid index shifting issues
        idxs.sort(reverse=True)

        for idx in idxs:

            if reference:

            # Merge the step with the next step
                self.merge_consecutive_steps(idx+1,idx+2)

            else:
                # Merge the step with the previous step
                self.merge_consecutive_steps(idx+2,idx+1)

        return None

    def read_sample_plate_info(self, files, names=None):
        """
        Read the sample plate information from the .fmf file.
        
        Parameters
        ----------
        files : str or list of str
            Path(s) to .fmf file(s) containing sample plate information.
        names : str or list of str, optional
            Name(s) to assign to the files. If None, file names are used.
            
        Returns
        -------
        None
            The method populates class attributes with sample plate data.
            
        Notes
        -----
        This method creates the following attributes:
            
        - sample_column : numpy.ndarray
          Array of sample column information (96 elements, one per well).
        - sample_row : numpy.ndarray
          Array of sample row information (96 elements, one per well).
        - sample_type : list
          List of sample types (96 elements, one per well).
        - sample_id : list
          List of sample IDs (96 elements, one per well).
        - sample_plate_loaded : bool
          Set to True if sample plate information is successfully loaded.
        - sample_conc : numpy.ndarray
          Array with the sample concentrations (96 elements, one per well).
        - sample_conc_labeled : list
          List with the sample concentrations labeled (96 elements, one per well).
        """

        if names is None:
            names = files

        if not isinstance(files, list):
            files = [files]
            names = [names]

        index = next((i for i, s in enumerate(names) if 'ExpMethod.fmf' in s), None)

        if index is None:

            self.sample_plate_loaded = False
            return None

        file = files[index]

        tree = elementTree.parse(file)
        root = tree.getroot()

        sample_types     = [x.text for x in root.findall(".//SampleType")]
        sample_locations = [x.text for x in root.findall(".//SampleLoc")]
        sample_ids       = [x.text for x in root.findall(".//SampleID")]

        sample_conc_molar    = np.array([float(x.text) for x in root.findall(".//SampleMolarConc")])
        sample_conc_mass     = np.array([float(x.text) for x in root.findall(".//SampleConc")])

        sample_conc          = sample_conc_mass

        sel_ids = ['SAMPLE' in s for s in sample_types]

        counter = 0
        for i in range(len(sample_conc)):

            if sel_ids[i]:

                if sample_conc_molar[counter] > 0:

                    sample_conc[i] = sample_conc_molar[counter]

                counter += 1

        conc_units       = [x.text for x in root.findall(".//ConcUnits")][0]
        molar_conc_units = [x.text for x in root.findall(".//MolarConcUnits")][0]

        factors = [factor_conc_to_micro[molar_conc_units] if 'SAMPLE' in st else factor_conc_to_micro[conc_units] for st in sample_types]

        sample_conc = sample_conc * np.array(factors)
        sample_conc = np.round(sample_conc, 5)

        sample_column = np.array([int(re.sub(r'\D', '', text)) for text in sample_locations])
        sample_row    = np.array([re.sub(r'\d+', '', text)     for text in sample_locations])

        self.sample_column = sample_column
        self.sample_row    = sample_row
        self.sample_type   = sample_types
        self.sample_id     = sample_ids

        self.sample_conc    = sample_conc

        sample_conc_labeled = [f"{x} µM" if t == 'KSAMPLE' and x >= 0 else f"{x} µg/ml" if t != 'KSAMPLE' and x >= 0 else '' for x, t in zip(sample_conc, sample_types)]

        self.sample_conc_labeled = sample_conc_labeled

        data_name     = [x.text for x in root.findall(".//DataName")]
        assay_time    = [x.text for x in root.findall(".//AssayTime")]

        steps_info_df      = pd.DataFrame({'Type':data_name,'Time':assay_time})

        data_name     = [x.text for x in root.findall(".//StepDataName")]
        data_col      = [x.text for x in root.findall(".//SampleCol")]

        steps_performed = pd.DataFrame({'#Step':np.arange(len(data_name))+1,'Type':data_name,'Column':data_col})

        steps_performed = pd.merge(steps_performed, steps_info_df, on='Type', how='left')

        self.steps_performed = steps_performed

        self.sample_plate_loaded = True

        return None

    def convert_to_numbers(self):
        """
        Convert the strings in the step info to numbers.
        
        Parameters
        ----------
        None
            
        Returns
        -------
        None
            The method modifies the step_info attribute in place.
            
        Notes
        -----
        This method processes the following entries in step_info:
        'Concentration', 'MolarConcentration', 'MolecularWeight', 'Temperature',
        'StartTime', 'AssayTime', 'FlowRate', 'ActualTime', 'CycleTime'.
        """

        # List of entries in step info
        entries = ['Concentration', 'MolarConcentration', 'MolecularWeight', 'Temperature', 'StartTime',
                   'AssayTime', 'FlowRate', 'ActualTime', 'CycleTime']

        for entry in entries:
            for sensor in range(len(self.fns)):
                # Do sanity check
                try:
                    self.step_info[sensor][entry] = np.array(self.step_info[sensor][entry], dtype=float)
                except:

                    print("Erroneous entry found for %s and sensor %i: %s" % (entry, sensor, self.step_info[sensor][entry]))
                    print("Will set it to -1. Needs to be corrected")
                    # Correct erroneous value
                    for i in range(len(self.step_info[sensor][entry])):
                        try:
                            float(self.step_info[sensor][entry][i])
                        except:
                            self.step_info[sensor][entry][i] = -1
                    self.step_info[sensor][entry] = np.array(self.step_info[sensor][entry], dtype=float)
        return None



import pytest
import numpy as np

from pykingenie.utils.signal_solution  import (
    signal_ode_conformational_selection_insolution,
    get_initial_concentration_conformational_selection
)

from pykingenie.utils.fitting_solution import fit_conformational_selection_solution

from src.pykingenie.utils.fitting_solution import find_initial_parameters_conformational_selection_solution


t = np.linspace(0, 0.5, 20)

E_tot = 1  # Total enzyme concentration
S_tot = np.logspace(-1.5, 1, 6)  # Total substrate concentrations

kon = 100
koff = 1
kc = 10
krev = 100

noise = np.random.default_rng(42).normal(0, 0.0001, len(t))

def generate_ys(E_tot,S_tot,kc,krev,kon,koff,signal_E1, signal_E2, signal_S, signal_E2S):

    ys = []

    for i in range(len(S_tot)):

        E1, E2 = get_initial_concentration_conformational_selection(E_tot,kc,krev)

        y0 = [E2, 0]

        y = signal_ode_conformational_selection_insolution(
            t, y0, kc,krev,kon,koff, E_tot, S_tot[i], 0,
            signal_E1, signal_E2, signal_S, signal_E2S)

        y += noise

        ys.append(y)

    return ys


def test_fit_cs_solution_fixed_constants_t0():

    signal_E = 0
    signal_S = 0
    signal_E2S = 20

    signal_lst = generate_ys(E_tot,S_tot[-3:],kc,krev,kon,koff,signal_E, signal_E, signal_S, signal_E2S)

    time_lst = [t + 0.03] * len(signal_lst)
    ligand_conc_lst = S_tot[-3:]
    protein_conc_lst = [E_tot] * len(signal_lst)

    initial_parameters = [20] + ([0.03] * len(time_lst))
    low_bounds = [10] + ([0.003] * len(time_lst))
    high_bounds = [30] + ([0.3] * len(time_lst))

    global_fit_params, _, _, params_names = fit_conformational_selection_solution(
        signal_lst=signal_lst,
        time_lst=time_lst,
        ligand_conc_lst=ligand_conc_lst,
        protein_conc_lst=protein_conc_lst,
        initial_parameters=initial_parameters,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        fixed_kon=True,
        fixed_koff=True,
        fixed_kc=True,
        fixed_krev=True,
        kon_value=kon,
        koff_value=koff,
        kc_value=kc,
        krev_value=krev,
        fit_signal_E=False,
        fit_signal_S=False,
        fit_signal_E2S=True,
        fixed_t0=False)

    expected_params = [signal_E2S] + ([0.03] * len(time_lst))  # t0 values should be close to the offset we added

    np.testing.assert_allclose(
        global_fit_params, expected_params, rtol=0.2,
        err_msg="The fitted parameters do not match the expected values."
    )

    expected_params_names = ["Signal of the complex"] + ['t0_' + str(i + 1) for i in range(len(time_lst))]

    assert params_names == expected_params_names, "The parameter names do not match the expected names."

def test_conformational_selection_insolution():

    signal_E1 = 0
    signal_S = 0
    signal_E2 = 0
    signal_E2S = 1

    signal_lst = generate_ys(E_tot,S_tot,kc,krev,kon,koff,signal_E1, signal_E2, signal_S, signal_E2S)

    time_lst = [t] * len(signal_lst)
    ligand_conc_lst = S_tot
    protein_conc_lst = [E_tot] * len(signal_lst)

    best_params = find_initial_parameters_conformational_selection_solution(
        signal_lst,
        time_lst,
        ligand_conc_lst,
        protein_conc_lst,
        np_linspace_low=0,
        np_linspace_high=3,
        np_linspace_num=4)

    # Fit using as initial parameters the best found parameters
    initial_parameters = np.array(best_params)
    low_bounds = best_params / 1e2
    high_bounds = best_params * 1e2

    global_fit_params, _, _, _ = fit_conformational_selection_solution(
        signal_lst, time_lst, ligand_conc_lst,
        protein_conc_lst, initial_parameters,
        low_bounds,high_bounds
    )

    expected_params = [signal_E2S, kon, koff, kc, krev]

    np.testing.assert_allclose(
        global_fit_params, expected_params, rtol=0.1,
        err_msg="The fitted parameters do not match the expected values.")


def test_fit_conformational_selection_solution_fixed_constants():

    signal_E1 = 0
    signal_E2 = 0
    signal_S = 0
    signal_E2S = 1

    signal_lst = generate_ys(E_tot,S_tot,kc,krev,kon,koff,signal_E1, signal_E2, signal_S, signal_E2S)

    time_lst = [t] * len(signal_lst)
    ligand_conc_lst = S_tot
    protein_conc_lst = [E_tot] * len(signal_lst)

    initial_parameters = [10]
    low_bounds = [1e-3]
    high_bounds = [1e3]

    global_fit_params, _, _, _ = fit_conformational_selection_solution(
        signal_lst=signal_lst,
        time_lst=time_lst,
        ligand_conc_lst=ligand_conc_lst,
        protein_conc_lst=protein_conc_lst,
        initial_parameters=initial_parameters,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        fixed_kon=True,
        fixed_koff=True,
        fixed_kc=True,
        fixed_krev=True,
        kon_value=kon,
        koff_value=koff,
        kc_value=kc,
        krev_value=krev,
        fit_signal_E=False,
        E1_equals_E2=True,
        fit_signal_S=False,
        fit_signal_E2S=True)

    expected_params = [signal_E2S]

    np.testing.assert_allclose(
        global_fit_params, expected_params, rtol=0.1,
        err_msg="The fitted parameters do not match the expected values.")


def test_fit_conf_sel_solution_fixed_constants_different_E_signal():

    signal_E1 = 1
    signal_E2 = 4
    signal_S = 0
    signal_E2S = 0

    signal_lst = generate_ys(E_tot,S_tot,kc,krev,kon,koff,signal_E1, signal_E2, signal_S, signal_E2S)

    time_lst = [t] * len(signal_lst)
    ligand_conc_lst = S_tot
    protein_conc_lst = [E_tot] * len(signal_lst)

    best_params = find_initial_parameters_conformational_selection_solution(
        signal_lst,
        time_lst,
        ligand_conc_lst,
        protein_conc_lst,
        np_linspace_low=0,
        np_linspace_high=3,
        np_linspace_num=4,
        fit_signal_E=True,
        E1_equals_E2=False,
        fit_signal_S=False,
        fit_signal_E2S=False,
        fixed_t0=True
        )

    # Fit using as initial parameters the best found parameters
    initial_parameters = np.array(best_params[:2])
    low_bounds = initial_parameters / 1e2
    high_bounds = initial_parameters * 1e2

    global_fit_params, _, _, params_names = fit_conformational_selection_solution(
        signal_lst=signal_lst,
        time_lst=time_lst,
        ligand_conc_lst=ligand_conc_lst,
        protein_conc_lst=protein_conc_lst,
        initial_parameters=initial_parameters,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        fixed_kon=True,
        fixed_koff=True,
        fixed_kc=True,
        fixed_krev=True,
        kon_value=kon,
        koff_value=koff,
        kc_value=kc,
        krev_value=krev,
        fit_signal_E=True,
        E1_equals_E2=False,
        fit_signal_S=False,
        fit_signal_E2S=False)

    expected_params = [signal_E1, signal_E2]

    np.testing.assert_allclose(
        global_fit_params, expected_params, rtol=0.1,
        err_msg="The fitted parameters do not match the expected values.")

    expected_params_names = [
        "Signal of the inactive protein (E1)",
        "Signal of the active protein (E2)"
    ]

    assert params_names == expected_params_names, "The parameter names do not match the expected names."


def test_fit_cs_solution_fixed_constants_different_signals():

    signal_E1 = 1
    signal_E2 = 1
    signal_S = 10
    signal_E2S = 5

    signal_lst = generate_ys(E_tot,S_tot,kc,krev,kon,koff,signal_E1, signal_E2, signal_S, signal_E2S)

    time_lst = [t] * len(signal_lst)
    ligand_conc_lst = S_tot
    protein_conc_lst = [E_tot] * len(signal_lst)

    best_params = find_initial_parameters_conformational_selection_solution(
        signal_lst,
        time_lst,
        ligand_conc_lst,
        protein_conc_lst,
        np_linspace_low=0,
        np_linspace_high=3,
        np_linspace_num=4,
        fit_signal_E=True,
        E1_equals_E2=True,
        fit_signal_S=True,
        fit_signal_E2S=True,
        fixed_t0=True
        )

    # Fit using as initial parameters the best found parameters
    initial_parameters = np.array(best_params[:3])
    low_bounds = initial_parameters / 1e2
    high_bounds = initial_parameters * 1e2

    global_fit_params, _, _, params_names = fit_conformational_selection_solution(
        signal_lst=signal_lst,
        time_lst=time_lst,
        ligand_conc_lst=ligand_conc_lst,
        protein_conc_lst=protein_conc_lst,
        initial_parameters=initial_parameters,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        fixed_kon=True,
        fixed_koff=True,
        fixed_kc=True,
        fixed_krev=True,
        kon_value=kon,
        koff_value=koff,
        kc_value=kc,
        krev_value=krev,
        fit_signal_E=True,
        E1_equals_E2=True,
        fit_signal_S=True,
        fit_signal_E2S=True)

    expected_params = [signal_E1, signal_S, signal_E2S]

    np.testing.assert_allclose(
        global_fit_params, expected_params, rtol=0.1,
        err_msg="The fitted parameters do not match the expected values."
    )

    expected_params_names = [
        "Signal of the free protein",
        "Signal of the free ligand",
        "Signal of the complex"]

    assert params_names == expected_params_names, "The parameter names do not match the expected names."

def test_fit_cs_solution_fixed_constants_different_signals_scale_factor():

    signal_E1 = 1
    signal_E2 = 1
    signal_S = 10
    signal_E2S = 5

    signal_lst = generate_ys(E_tot,S_tot,kc,krev,kon,koff,signal_E1, signal_E2, signal_S, signal_E2S)

    time_lst = [t] * len(signal_lst)
    ligand_conc_lst = S_tot
    protein_conc_lst = [E_tot] * len(signal_lst)

    best_params = find_initial_parameters_conformational_selection_solution(
        signal_lst,
        time_lst,
        ligand_conc_lst,
        protein_conc_lst,
        np_linspace_low=0,
        np_linspace_high=3,
        np_linspace_num=4,
        fit_signal_E=True,
        E1_equals_E2=True,
        fit_signal_S=True,
        fit_signal_E2S=True,
        fixed_t0=True,
        fit_scale_factor=True
        )

    # Fit using as initial parameters the best found parameters
    initial_parameters = np.array(best_params[:3])
    low_bounds = initial_parameters / 1e2
    high_bounds = initial_parameters * 1e2

    initial_parameters = np.concatenate((initial_parameters, [1] * len(time_lst)))
    low_bounds = np.concatenate((low_bounds, [0.7] * len(time_lst)))
    high_bounds = np.concatenate((high_bounds, [1.3] * len(time_lst)))

    global_fit_params, _, _, params_names = fit_conformational_selection_solution(
        signal_lst=signal_lst,
        time_lst=time_lst,
        ligand_conc_lst=ligand_conc_lst,
        protein_conc_lst=protein_conc_lst,
        initial_parameters=initial_parameters,
        low_bounds=low_bounds,
        high_bounds=high_bounds,
        fixed_kon=True,
        fixed_koff=True,
        fixed_kc=True,
        fixed_krev=True,
        kon_value=kon,
        koff_value=koff,
        kc_value=kc,
        krev_value=krev,
        fit_signal_E=True,
        E1_equals_E2=True,
        fit_signal_S=True,
        fit_signal_E2S=True,
        fit_scale_factor=True,
        fixed_t0=True
        )

    expected_params = [signal_E1, signal_S, signal_E2S] + ([1] * len(time_lst))

    np.testing.assert_allclose(
        global_fit_params, expected_params, rtol=0.1,
        err_msg="The fitted parameters do not match the expected values."
    )

    expected_params_names = [
        "Signal of the free protein",
        "Signal of the free ligand",
        "Signal of the complex",
        "scale_factor_1",
        "scale_factor_2",
        "scale_factor_3",
        "scale_factor_4",
        "scale_factor_5",
        "scale_factor_6"
    ]

    assert params_names == expected_params_names, "The parameter names do not match the expected names."


import numpy as np
import pandas as pd

from scipy import stats

___all__ = ['single_exponential',
           'double_exponential',
           'median_filter',
           'rss_p',
           'get_rss',
           'get_desired_rss']

def single_exponential(t, a0, a1, kobs):
    """
    Single exponential function for fitting.

    Parameters
    ----------
    t : np.ndarray
        Time values.
    a0 : float
        Offset.
    a1 : float
        Amplitude.
    kobs : float
        Observed rate constant.

    Returns
    -------
    np.ndarray
        Computed values of the single exponential function.
    """
    return a0 + a1 * np.exp(-kobs * t)

def double_exponential(t, a0, a1, kobs1, a2, kobs2):
    """
    Double exponential function for fitting.

    Parameters
    ----------
    t : np.ndarray
        Time values.
    a0 : float
        Offset.
    a1 : float
        Amplitude of the first exponential.
    kobs1 : float
        Observed rate constant of the first exponential.
    a2 : float
        Amplitude of the second exponential.
    kobs2 : float
        Observed rate constant of the second exponential.

    Returns
    -------
    np.ndarray
        Computed values of the double exponential function.
    """
    return a0 + a1 * np.exp(-kobs1 * t) + a2 * np.exp(-kobs2 * t)

def median_filter(y, x, rolling_window):
    """
    Compute the median filter of the input signal using a rolling window.

    The x vector is converted into an integer vector and then into a time variable to take advantage of pandas rolling().median().
    Returns the y vector passed through the median filter.

    Parameters
    ----------
    y : np.ndarray
        Input signal values.
    x : np.ndarray
        Time or index values corresponding to the signal.
    rolling_window : int or float
        Size of the median filter window.

    Returns
    -------
    np.ndarray
        Filtered signal.
    """

    # Check the average time step in the x vector
    dx = np.min(np.diff(x))

    scaling_factor = 1
    if dx <= 1:
        scaling_factor = 1/dx * 10 # Rescale 
    
    temp_vec     =  np.multiply(x,scaling_factor).astype(int)
    series       =  pd.Series(y,index=temp_vec,dtype=float)
    series.index =  pd.to_datetime(series.index,unit='s') # Convert to datetime index - allows rolling window

    roll_window  = str(int(rolling_window*scaling_factor))+"s"

    y_filt = series.rolling(
        window=roll_window).median().to_numpy()

    return y_filt

def rss_p(rrs0, n, p, alfa):
    """
    Compute the desired residual sum of squares for a 1-alpha confidence interval.

    Given the residuals of the best fitted model, this is used to compute asymmetric confidence intervals for the fitted parameters.

    Parameters
    ----------
    rrs0 : float
        Residual sum of squares of the model with the best fit.
    n : int
        Number of data points.
    p : int
        Number of parameters.
    alfa : float
        Desired confidence interval (1 - alpha).

    Returns
    -------
    float
        Residual sum of squares for the desired confidence interval.
    """

    critical_value = stats.f.ppf(q=1 - alfa, dfn=1, dfd=n - p)

    return rrs0 * (1 + critical_value / (n - p))

def get_rss(y, y_fit):
    """
    Compute the residual sum of squares.

    Parameters
    ----------
    y : np.ndarray
        Observed values.
    y_fit : np.ndarray
        Fitted values.

    Returns
    -------
    float
        Residual sum of squares.
    """

    residuals = y - y_fit
    rss       = np.sum(residuals ** 2)

    return rss

def get_desired_rss(y, y_fit, n, p, alpha=0.05):
    """
    Find the residual sum of squares required for a 1-alpha confidence interval.

    Given the observed and fitted data, computes the RSS for the desired confidence interval.

    Parameters
    ----------
    y : np.ndarray
        Observed values.
    y_fit : np.ndarray
        Fitted values.
    n : int
        Number of data points.
    p : int
        Number of parameters.
    alpha : float, optional
        Desired confidence interval (default is 0.05).

    Returns
    -------
    float
        Residual sum of squares for the desired confidence interval.
    """

    rss = get_rss(y, y_fit)

    return rss_p(rss, n, p, alpha)



import pandas as pd

class KineticsFitterGeneral:

    """
    A class used to handle kinetics data both for solution-based and surface-based experiments.
    """

    def __init__(self):
        """
        Initialize the KineticsFitterGeneral instance.

        Attributes
        ----------
        signal_ss_fit : list of numpy.ndarray or None
            Fitted steady state signals.
        signal_assoc_fit : list of numpy.ndarray or None
            Fitted association kinetics signals.
        signal_disso_fit : list of numpy.ndarray or None
            Fitted dissociation kinetics signals.
        fit_params_kinetics : pandas.DataFrame or None
            Fitted parameters for the association / dissociation kinetics.
        fit_params_ss : pandas.DataFrame or None
            Fitted parameters for the steady state.
        fitted_params_boundaries : pandas.DataFrame or None
            DataFrame with fitting boundaries and fitted parameters.
        params : numpy.ndarray or None
            Parameters used for fitting.
        low_bounds : numpy.ndarray or None
            Lower bounds for fitting parameters.
        high_bounds : numpy.ndarray or None
            Upper bounds for fitting parameters.
        """

        self.signal_ss_fit = None  # Steady state fitted signal
        self.signal_assoc_fit = None  # Association kinetics fitted signal
        self.signal_disso_fit = None  # Dissociation kinetics fitted signal
        self.fit_params_kinetics = None  # Fitted parameters for the association / dissociation kinetics
        self.fit_params_ss = None  # Values of the fitted parameters - steady state
        self.fitted_params_boundaries = None  # DataFrame with fitting boundaries and fitted parameters
        self.params = None  # Parameters used for fitting
        self.low_bounds = None  # Lower bounds for fitting parameters
        self.high_bounds = None  # Upper bounds for fitting parameters

    def clear_fittings(self):
        """
        Clear all fitting results.

        Resets all fitted signal arrays and parameter dataframes to None.

        Returns
        -------
        None
        """
        self.signal_ss_fit = None  # Steady state fitted signal
        self.signal_assoc_fit = None  # Association kinetics fitted signal
        self.signal_disso_fit = None  # Association kinetics fitted signal
        self.fit_params_kinetics = None  # Fitted parameters for the association / dissociation kinetics
        self.fit_params_ss = None  # Values of the fitted parameters - steady state

        return None

    def create_fitting_bounds_table(self):
        """
        Create a dataframe with the fitting bounds and the fitted parameters.

        Uses the class attributes params, low_bounds and high_bounds to create
        a DataFrame with fitting boundaries.

        Returns
        -------
        None
            Updates the fitted_params_boundaries attribute with a DataFrame.
        """
        df = pd.DataFrame({
            'Fitted_parameter_value': self.params,
            'Lower_limit_for_fitting': self.low_bounds,
            'Upper_limit_for_fitting': self.high_bounds
        })

        self.fitted_params_boundaries = df

        return None